from pathlib import Path
import os
import shutil
import signal
from typing import Union


class WorkDirContext:
    def __init__(self, path: Union[Path, str]):
        self.path = Path(path)
        self.origin = Path(os.getcwd()).absolute()

    def __enter__(self):
        os.chdir(self.path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.origin)


def force_mkdir(path: Path):
    path = Path(path).expanduser().absolute()
    if path.exists():
        shutil.rmtree(path, ignore_errors=True)

    try:
        os.mkdir(path)
    except Exception as e:
        print(f"Failed to make directory because {e}")


class timeout:
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type_, value, traceback):
        signal.alarm(0)




from pathlib import Path
from typing import Dict
import logging

from cfgutils.sorting import cfg_root_node
import toml
import networkx as nx


l = logging.getLogger(__name__)

def read_line_maps(maps_path: Path, key_is_func=True, value_is_set_type=True):
    with open(maps_path, "r") as fp:
        line_maps = toml.load(fp)

    # convert keys to int and values to set
    if key_is_func:
        for func in line_maps:
            for linenum in list(line_maps[func]):
                if value_is_set_type:
                    line_maps[func][int(linenum)] = set(line_maps[func][linenum])
                else:
                    line_maps[func][int(linenum)] = line_maps[func][linenum]
                del line_maps[func][linenum]
    else:
        for addr in list(line_maps.keys()):
            line_maps[int(addr)] = set(line_maps[addr])
            del line_maps[addr]

    return line_maps


def correct_decompiler_mappings(dec_cfg, dec_line_to_addr_map):
    node = cfg_root_node(dec_cfg)
    if node is None:
        raise Exception("Could not find root node!")

    base = node.addr - 1
    corrected_line_nums = {k+base: v for k, v in dec_line_to_addr_map.items()}
    line_nums = corrected_line_nums.keys()
    first_ln = min(line_nums)
    last_ln = max(line_nums)

    for ln in range(first_ln, last_ln):
        val = corrected_line_nums.get(ln, None)
        if val is not None:
            continue

        # search for a value expanding pos and negative equally until
        # another line have a value that can be used
        reach = 1
        is_neg = True
        while reach < last_ln:
            curr_reach = 0 - reach if is_neg else reach
            val = corrected_line_nums.get(ln + curr_reach, None)
            # leave when we find a good val
            if val is not None:
                corrected_line_nums[ln] = val
                break

            reach += 1
            is_neg ^= True

    return corrected_line_nums


def correct_source_cfg_addrs(cfgs: Dict[str, nx.DiGraph], line_map_file: Path):
    line_map_file = Path(line_map_file).absolute()
    func_correct_line_map = read_line_maps(line_map_file, value_is_set_type=False)

    new_cfgs = {}
    for cfg_name, cfg in cfgs.items():
        # skip cfgs without mappings
        line_to_correct_map = func_correct_line_map.get(cfg_name, None)
        if line_to_correct_map is None:
            continue

        nodes_map = {}
        new_graph = nx.DiGraph()
        for node in cfg:
            node_copy = node.copy()
            new_addr = line_to_correct_map.get(node.addr, None)
            if new_addr is not None:
                node_copy.addr = new_addr
            else:
                l.warning(f"Parsed a block: {node} without a mapping for lineaddrs!")

            for stmt in node_copy.statements:
                new_addr = line_to_correct_map.get(stmt.source_line_number, None)
                if new_addr is None:
                    l.warning(f"Parsed a block statement: {stmt} without a mapping for lineaddrs!")
                    continue

                stmt.source_line_number = new_addr

            nodes_map[node] = node_copy

        for node in nodes_map.values():
            new_graph.add_node(node, node=node)

        for src, dst in cfg.edges:
            new_graph.add_edge(nodes_map[src], nodes_map[dst], src=nodes_map[src], dst=nodes_map[dst])

        new_cfgs[cfg_name] = new_graph

    return new_cfgs




import sys
import unittest
from pathlib import Path

from pyjoern import fast_cfgs_from_source, parse_source, JoernClient, JoernServer

TEST_SOURCE_DIR = Path(__file__).absolute().parent / "source"


class TestParsing(unittest.TestCase):
    def test_parser(self):
        functions = parse_source(TEST_SOURCE_DIR / "simple.c")
        assert len(functions) == 2

        schedule_job = functions.get("schedule_job", None)
        assert schedule_job is not None
        assert schedule_job.name == "schedule_job"
        assert schedule_job.start_line == 1
        assert schedule_job.end_line == 19
        assert len(schedule_job.gotos) == 1
        assert schedule_job.cfg is not None

    def test_fast_cfg(self):
        cfg = fast_cfgs_from_source(TEST_SOURCE_DIR / "simple.c")["main"]
        assert cfg is not None, "CFG did not actually parse"


class TestJoernServer(unittest.TestCase):
    def test_server(self):
        with JoernServer() as server:
            assert server is not None

            client = JoernClient(TEST_SOURCE_DIR / "simple.c", port=server.port)
            assert client is not None

            funcs = client.functions
            assert 'main' in funcs


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from textwrap import dedent
from typing import Union, List

from cfgutils.data import GenericBlock
from .statement import MergedRegionStart, Nop


class Block(GenericBlock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_merged_node = kwargs.get("is_merged_node", False)
        self._is_entrypoint = kwargs.get("is_entrypoint", False)
        self._is_exitpoint = kwargs.get("is_exitpoint", False)

    def __repr__(self):
        return f"<Block: {self.addr}{self._idx_str}, {len(self.statements)} statements>"

    def __str__(self):
        output = f"{self.addr}{self._idx_str}:\n"
        for line in self.statements:
            output += f"{line}\n"

        return output

    def contains_addr(self, addr):
        if self.addr == addr:
            return True
        
        for stmt in self.statements:
            if stmt.source_line_number == addr:
                return True

        return False

    @property
    def is_merged_node(self):
        if self._is_merged_node:
            return True

        return self.statements and isinstance(self.statements[0], MergedRegionStart)

    @is_merged_node.setter
    def is_merged_node(self, data):
        self._is_merged_node = data

    @property
    def is_entrypoint(self):
        if self._is_entrypoint:
            return True

        if not self.statements:
            return False

        first_stmt = self.statements[0]
        if isinstance(first_stmt, MergedRegionStart):
            for stmt in self.statements:
                if isinstance(stmt, Nop) and stmt.type == Nop.FUNC_START:
                    return True
        elif isinstance(first_stmt, Nop) and first_stmt.type == Nop.FUNC_START:
            return True
        else:
            return False

    @is_entrypoint.setter
    def is_entrypoint(self, data):
        self._is_entrypoint = data

    @property
    def is_exitpoint(self):
        if self._is_exitpoint:
            return True

        if not self.statements:
            return False

        last_stmt = self.statements[-1]
        first_stmt = self.statements[0]
        if isinstance(last_stmt, Nop) and last_stmt.type == Nop.FUNC_END:
            return True
        elif isinstance(first_stmt, MergedRegionStart):
            for stmt in self.statements:
                if isinstance(stmt, Nop) and stmt.type == Nop.FUNC_END:
                    return True
        else:
            return False

    @is_exitpoint.setter
    def is_exitpoint(self, data):
        self._is_exitpoint = data

    @classmethod
    def merge_many_blocks(cls, start_addr, nodes: List["GenericBlock"]):
        node_count = 0
        stmts = list()
        for node in nodes:
            for stmt in node.statements:
                node_count += stmt.total_nodes if isinstance(stmts, MergedRegionStart) else 1
                stmts.append(stmt)

        return Block(
            start_addr,
            statements=[MergedRegionStart(source_line_number=start_addr, total_nodes=node_count)] + stmts,
            is_merged_node=True,
        )




from pathlib import Path

import networkx as nx

from ..cfg import parse_dot_cfg_string, normalize_cfg


class Function:
    def __init__(
        self,
        name: str = None,
        return_type: str | None = None,
        calls: list[str] | None = None,
        macro_count: int | None = None,
        control_structures: list[str] | None = None,
        gotos: list[str] | None = None,
        fullname: str | None = None,
        filename: str | None = None,
        start_line: int | None = None,
        end_line: int | None = None,
        signature: str | None = None,
        cfg: nx.DiGraph | str | list | None = None,
    ):
        self.name = name
        self.return_type = return_type
        self.calls = calls or []
        self.macro_count = macro_count or 0
        self.control_structures = control_structures or []
        self.gotos = gotos or []
        self.fullname = fullname
        self.filename: Path | None = Path(filename)
        self.start_line = start_line
        self.end_line = end_line
        self.signature = signature
        self.cfg = self._parse_dot_cfg_string(cfg) if isinstance(cfg, (str, list)) else cfg

    def __str__(self):
        return f"<Function {self.name} {self.signature} {self.start_line}:{self.end_line}>"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @staticmethod
    def _parse_dot_cfg_string(cfg_data: str | list) -> nx.DiGraph:
        cfg_str = cfg_data if isinstance(cfg_data, str) else cfg_data[0]
        parsed_cfg = parse_dot_cfg_string(cfg_str)
        if parsed_cfg is not None:
            parsed_cfg = normalize_cfg(parsed_cfg, lift_cfg=True, supergraph=True)

        return parsed_cfg

    @staticmethod
    def from_many(data: list[dict]) -> dict[tuple[str, str], "Function"]:
        functions = [Function(**d) for d in data]
        blacklisted_names = [
            "<", "+", "*", "(", ">", "JUMPOUT", "__builtin_unreachable"
        ]
        # func_by_name[(name, filename)] = function
        func_by_name = {}
        for f in functions:
            # remove blacklisted
            if not f.name or not f.filename or any(f.name.startswith(bl) for bl in blacklisted_names):
                continue

            # removed errored (or empty) CFGs
            if not f.cfg or not len(f.cfg.nodes):
                continue

            # normalize out declarations
            prev_func = func_by_name.get((f.name, str(f.filename)), None)
            if prev_func is not None and prev_func.cfg is not None and len(f.cfg.nodes) < len(prev_func.cfg.nodes):
                continue

            func_by_name[(f.name, str(f.filename))] = f

        return func_by_name



#
# JIL:
# Joern Intermediate Language
#

from copy import deepcopy


class Statement:
    def __init__(self, raw_text, source_line_number=None, **kwargs):
        self.raw_text = raw_text
        self.source_line_number = source_line_number

    @classmethod
    def op_seperator(cls, sub_type):
        return None

    def __str__(self):
        return self.raw_text

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.__str__()}>"

    def copy(self):
        # TODO make this better in the future!
        return deepcopy(self)


class Assignment(Statement):
    def __init__(self, raw_text, src, dst, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.src = src
        self.dst = dst

    @classmethod
    def op_seperator(cls, sub_type):
        return "="

    def __str__(self):
        return f"{self.src} = {self.dst}"


class Return(Statement):
    def __init__(self, raw_text, ret, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.ret = ret

    @classmethod
    def op_seperator(cls, sub_type):
        return "return"

    def __str__(self):
        return f"return {self.ret}"


class Call(Statement):
    def __init__(self, raw_text, func, args, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.func = func
        self.args = args

    @classmethod
    def op_seperator(cls, sub_type):
        return "return"

    def __str__(self):
        return f"{self.func}({''.join(self.args)})"


class Compare(Statement):
    EQ = 0
    GTE = 1
    GT = 2
    LTE = 3
    LT = 4
    NE = 5

    def __init__(self, raw_text, type_, arg1, arg2, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.type = type_
        self.arg1 = arg1
        self.arg2 = arg2

    def _pretty_seperator(self):
        op_map = {
            self.EQ: "==",
            self.GTE: ">=",
            self.GT: ">",
            self.LT: "<",
            self.LTE: "<=",
            self.NE: "!=",
        }

        return op_map[self.type]

    @classmethod
    def op_seperator(cls, sub_type):
        op_map = {
            cls.EQ: "==",
            cls.GTE: "&gt;=",
            cls.GT: "&gt;",
            cls.LT: "&lt;",
            cls.LTE: "&lt;=",
            cls.NE: "!="
        }

        return op_map.get(sub_type, None)

    def __str__(self):
        return f"{self.arg1} {self._pretty_seperator()} {self.arg2}"


class Ternary(Statement):
    def __init__(self, raw_text, cond, true, false, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.cond = cond
        self.true = true
        self.false = false

    @classmethod
    def op_seperator(cls, sub_type):
        return ""

    def __str__(self):
        return f"{self.cond} ? {self.true} : {self.false}"


class UnaryOp(Statement):
    pass


class BinOp(Statement):
    SUB = 0
    ADD = 1
    AND = 2
    OR = 3

    def __init__(self, raw_text, type_, arg1, arg2, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.type = type_
        self.arg1 = arg1
        self.arg2 = arg2

    def _pretty_seperator(self):
        op_map = {
            self.SUB: "-",
            self.ADD: "+",
            self.AND: "&&",
            self.OR: "||",
        }

        return op_map[self.type]

    @classmethod
    def op_seperator(cls, sub_type):
        op_map = {
            cls.SUB: "minus",
            cls.ADD: "plus",
            cls.AND: "logicalAnd",
            cls.OR: "logicalOr",
        }

        return op_map.get(sub_type, None)

    def __str__(self):
        return f"{self.arg1} {self._pretty_seperator()} {self.arg2}"


class Nop(Statement):
    FUNC_START = 0
    FUNC_END = 1
    NOP = 2

    str_map = {
        FUNC_START: "FUNCTION_START",
        FUNC_END: "FUNCTION_END",
        NOP: "NOP"
    }

    def __init__(self, raw_text, type_, **kwargs):
        super().__init__(raw_text, **kwargs)
        self.type = type_

    def __str__(self):
        return f"{self.str_map.get(self.type, None)}"


class UnknownStmt(Statement):
    def __str__(self):
        return f"<{self.__class__.__name__}: {self.raw_text}>"

    def __repr__(self):
        return self.__str__()


class UnsupportedStmt(Statement):
    def __str__(self):
        return f"<{self.__class__.__name__}: {self.raw_text}>"

    def __repr__(self):
        return self.__str__()


class MergedRegionStart(Statement):
    def __init__(self, source_line_number=None, total_nodes=0, **kwargs):
        super().__init__("", source_line_number=source_line_number)
        self.total_nodes = total_nodes

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.source_line_number}, {self.total_nodes} Nodes>"



import argparse


def main():
    parser = argparse.ArgumentParser(
        description="The command line tool for PyJoern. Do things like initializing Joern and running queries."
    )
    parser.add_argument(
        "--install", action="store_true", help="Install Joern backend and dependencies."
    )
    args = parser.parse_args()

    if args.install:
        import pyjoern



if __name__ == "__main__":
    main()



__version__ = "4.0.61.2"

import hashlib
import os
from pathlib import Path
import importlib.resources
import subprocess
import urllib.request
import math

from tqdm import tqdm

# initialize logging for the entire project
import logging
logging.getLogger("pyjoern").addHandler(logging.NullHandler())
from .logger import Loggers
loggers = Loggers()
del Loggers

_l = logging.getLogger(__name__)

JOERN_VERSION = "v4.0.61"
JOERN_BIN_DIR_PATH = Path(Path(str(importlib.resources.files("pyjoern"))) / "bin/joern-cli").absolute()
JOERN_SERVER_PATH = JOERN_BIN_DIR_PATH / "joern"
JOERN_EXPORT_PATH = JOERN_BIN_DIR_PATH / "joern-export"
JOERN_PARSE_PATH = JOERN_BIN_DIR_PATH / "joern-parse"
# must update both of these on supported Joern backend update
JOERN_ZIP_HASH = "fa40c38554d26942abec8dd5de9933d48269ccaa1e0271dfcd134801bfc4bc4f988d6b629f2bacc75a0fecc97a31393cfe46e8effb01692c1a1994ddcffa7cf4"
# must be imported after defining project wide constants
from .client import JoernClient
from .server import JoernServer
from .cfg import fast_cfgs_from_source
from .parsing import parse_source


#
# Helper code for downloading Joern server backend
#

def _download_and_save_joern_zip(save_location: Path, verify=True) -> Path:
    # XXX: hacked code for non-ssl verification
    import ssl

    ssl._create_default_https_context = ssl._create_unverified_context

    url = f"https://github.com/joernio/joern/releases/download/{JOERN_VERSION}/joern-cli.zip"
    with urllib.request.urlopen(url) as response:
        total_size = response.length
        if response.status != 200:
            raise Exception(f"HTTP error {response.status}: {response.reason}")

        hasher = hashlib.sha512()
        chunk_size = 8192
        mb_size = int(total_size / 1000000)
        with open(save_location, 'wb') as f:
            for _ in tqdm(range(math.ceil(total_size / chunk_size)), desc=f"Downloading Joern bytes (~{mb_size} MB)..."):
                chunk = response.read(chunk_size)
                hasher.update(chunk)
                if not chunk:
                    break

                f.write(chunk)

        # hash for extra security
        download_hash = hasher.hexdigest()
        if verify and download_hash != JOERN_ZIP_HASH:
            raise Exception(f"Joern files corrupted in download: {download_hash} != {JOERN_ZIP_HASH}")

    return save_location


def _download_joern():
    joern_binary = JOERN_BIN_DIR_PATH / "joern"
    if joern_binary.exists():
        return

    # download joern
    if not JOERN_BIN_DIR_PATH.parent.exists():
        os.mkdir(JOERN_BIN_DIR_PATH.parent)
    joern_zip_file = _download_and_save_joern_zip(JOERN_BIN_DIR_PATH.parent / "joern-cli.zip", verify=True)
    # unzip joern
    subprocess.run(["unzip", str(joern_zip_file)], cwd=str(JOERN_BIN_DIR_PATH.parent), capture_output=True, check=True)
    # remove zip file
    joern_zip_file.unlink()

    if not joern_binary.exists():
        raise Exception("Failed to download Joern!")


if not JOERN_BIN_DIR_PATH.exists():
    _l.info(f"Joern binaries are not installed on your system, downloading them now...")
    _download_joern()
    if not JOERN_BIN_DIR_PATH.exists():
        raise Exception("Failed to download Joern on startup!")



import os
import re
import logging
from typing import Set
from pathlib import Path

from cpgqls_client import CPGQLSClient, import_code_query
import networkx as nx

l = logging.getLogger(__name__)


class JoernClient:
    def __init__(self, target_file, ip="localhost", port=9000, bin_name=None):
        self.ip = ip
        self.port = port
        self.target_file = Path(target_file).absolute()
        self.bin_name = bin_name or ""

        # connect and import target file
        self.client = CPGQLSClient(f"{ip}:{port}")
        self.client.execute(import_code_query(f"{self.target_file}", f"{os.path.basename(self.target_file)}"))
        self.functions: Set[str] = self._get_functions_with_code()

        # cache
        self.cfg_cache = {}

    #
    # Public API
    #
    
    def function_line_numbers(self):
        out = self._exec_list_cmd("cpg.method.filter(node => node.lineNumber!=None&&node.lineNumberEnd!=None).map(node => (node.name, node.lineNumber.last, node.lineNumberEnd.last)).l")
        if not out:
            return {}

        if not isinstance(out, (list, tuple)):
            return {}

        line_nums_by_func = {}
        for data in out:
            if not isinstance(data, (list, tuple)) or len(data) != 3:
                continue

            fn, ln_start, ln_end = data
            if fn not in self.functions:
                continue

            line_nums_by_func[fn] = (ln_start, ln_end)

        return line_nums_by_func

    def function_headers(self):
        headers = {}
        for func in self.functions:
            out = self._exec_list_cmd(f'cpg.method.filter(_.name=="{func}").code.l')
            if not out:
                continue

            if not isinstance(out, (list, tuple)):
                continue

            headers[func] = out[0]

        return headers

    def function_code(self, full_funcname):
        out = self._exec_list_cmd(f'cpg.method.filter(_.fullName=="{full_funcname}").code.l')
        if not out:
            return None

        if not isinstance(out, (list, tuple)):
            return 
  
        return out[0]
    
    def functions_with_gotos(self):
        out1 = self._exec_list_cmd(f'cpg.goto.method.name.l')
        if not out1:
            return []

        if isinstance(out1, (list, tuple)) and len(out1) > 0:
            return out1

        return []
        
    def functions_with_switches(self):
        out1 = self._exec_list_cmd(f'cpg.switchBlock.method.name.l')
        if not out1:
            return []

        if isinstance(out1, (list, tuple)) and len(out1) > 0:
            return out1

        return []

    def get_func_cfg(self, func_name):
        if func_name in self.cfg_cache:
            return self.cfg_cache[func_name]

        cfg = self._dump_func_cfg(func_name)
        self.cfg_cache[func_name] = cfg

        return cfg

    def get_func_loc(self, func_name):
        """
        Gets a functions lines of code (loc) count.
        """
        if func_name not in self.functions:
            return None

        out1 = self._exec_list_cmd(f'cpg.method.filter(_.name=="{func_name}").lineNumber.l')
        if out1 is None:
            return None

        out2 = self._exec_list_cmd(f'cpg.method.filter(_.name=="{func_name}").lineNumberEnd.l')
        if out2 is None:
            return None

        start_line = out1[-1] if not isinstance(out1, int) else out1
        end_line = out2[-1] if not isinstance(out2, int) else out2

        try:
            val = int(end_line) - int(start_line)
        except Exception:
            return None
        return val
    
    def function_ternary_counts(self):
        count = {}
        for func_name in self.functions:
            out = self._exec_int_cmd(f'cpg.method.filter(_.name=="{func_name}").call.filter(_.name=="<operator>.conditional").filter(_.code.contains("__builtin_unreachable")==false).size')
            if out is None:
                continue
            count[func_name] = out 

        return count
            

    def count_gotos(self, func_name):
        if func_name not in self.functions:
            return None

        out = self._exec_int_cmd(f'cpg.method.filter(_.name=="{func_name}").goto.size')
        if out is None:
            return None

        return out

    def count_if_levels(self, func_name):
        """
        Returns the max of if-nesting level

        """
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{func_name}").controlStructure.controlStructureType("IF").depth(_.isControlStructure).l'
        )
        if out is None or (isinstance(out, (list, tuple)) and len(out) == 0):
            return None

        # if_count = out if isinstance(out, int) else sum(out) // len(out)
        if_count = out if isinstance(out, int) else max(out)
        return if_count

    def func_calls_in_func(self, func_name):
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{func_name}").call.name.l'
        )
        if out is None:
            return None

        out = [out] if isinstance(out, str) else out
        # good_funcs = self._filter_blacklisted(out)
        # return good_funcs
        return out
    
    def func_calls_in_func_with_fullname(self, func_name):
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{func_name}").call.methodFullName.l'
        )
        if out is None:
            return None

        out = [out] if isinstance(out, str) else out
        # good_funcs = self._filter_blacklisted(out)
        # return good_funcs
        return out
    
    def get_control_structure_conditions(self, func_name):
        if func_name not in self.functions:
            return None

        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{func_name}").controlStructure.condition.code.l'
        )
        if not isinstance(out, (tuple, list)):
            return None

        return list(out)

    def get_filename(self, func_name):
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{func_name}").filename.l'
        )
        if out is None:
            return None
        return out[0]
    
    def get_full_funcname(self, func_name):
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{func_name}").fullName.l'
        )
        if out is None:
            return None
        return out[0]

    def get_function_details(self, funcname):
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{funcname}").filter(node => node.lineNumber!=None&&node.lineNumberEnd!=None&&node.body.typeFullName!="<empty>").map(x => (x.fullName, x.filename, x.lineNumber.getOrElse(None), x.lineNumberEnd.getOrElse(None), x.signature, x.name, x.signature)).l'
        )
        if not isinstance(out, (tuple, list)):
            function_details = []
            return function_details
        out = out if isinstance(out, tuple) and isinstance(out[0], tuple) else (out,)
        
        function_details = []
        
        for fullfuncname, filename, start_line, end_line, func_return_type, name, signature in out:
            if start_line is not None and end_line is not None: 
                function_details.append({
                    'fullfuncname': fullfuncname, 
                    'filename': filename, 
                    'start_line': start_line, 
                    'end_line': end_line,
                    'func_return_type': func_return_type,
                    'name': name, # We need this fied cuz we need to compare it with fullfuncname
                    'signature': signature # 
                })

        return function_details

    def func_arguments(self, funcname):
        try:
            out = self._exec_list_cmd(
                f'cpg.method.filter(_.name=="{funcname}").flatMap(m => m.parameter.map(x => (m.filename, x.name, x.code, x.lineNumber.getOrElse(None), x.columnNumber.getOrElse(None)))).l'
            )
            if not out:
                return {}
            out = out if isinstance(out, tuple) and isinstance(out[0], tuple) else (out,)
                
            argnames_list = {}
            for filename, argname, code, line, col in out:
                if line is None:
                    continue
                if filename not in argnames_list:
                    argnames_list[filename] = {}
                argnames_list[filename][argname] = {'code': code, 'line': line, 'column': col}
            return argnames_list
        
        # TODO: why do I get extra args with no linenumber?
        except Exception as e:
            l.warning(f"Error occurred doing JOERN eval for {funcname} while collecting args because {e}")
            return {}
        
    def isMacro(self,fname):
        out = self.client.execute(
            f'cpg.method.filter(node => node.lineNumber!=None&&node.lineNumberEnd!=None&&node.body.typeFullName!="<empty>").filter(_.name=="{fname}").where(_.body.typeFullName("ANY")).size'
        )
        out = int(out['stdout'].split("Int = ")[1])
        if out != 0:
            return True
        else:
            return False

    def global_variables(self):
        out = self._exec_list_cmd(
        f'cpg.local.where(_.method.filter(_.name=="<global>")).map(x=> (x.name, x.lineNumber.getOrElse(None), x.columnNumber.getOrElse(None))).l'
        )

        if not out:
            return {}
        out = out if isinstance(out, tuple) and isinstance(out[0], tuple) else (out,)
        
        vars = {varname: {'line': line, 'column': col} for varname, line, col in out}
        return vars

    def local_variables(self, funcname):
        out = self._exec_list_cmd(
            f'cpg.method.filter(_.name=="{funcname}").flatMap(m => m.local.map(x => (m.filename, x.name, x.lineNumber.getOrElse(None), x.columnNumber.getOrElse(None)))).l'
        )

        if not out:
            return {}
        out = out if isinstance(out, tuple) and isinstance(out[0], tuple) else (out,)
        
        vars = {}
        for filename, varname, line, col in out:
            if line is None:
                continue
            if filename not in vars:
                vars[filename] = {}
            vars[filename][varname] = {'line': line, 'column': col}

        return vars
    
    # TODO: update this!
    def graph_edges_with_properties(self, code, property):
        out = self._list_parsing_regex(
            f'cpg.graph.E.map(edge=>(edge.inNode.id, edge.outNode.id, edge.label, edge.propertiesMap.get("{property}"))).l')
        
        if out is None:
            return []
        return out
    
    def graph_nodes(self, code):
        out = self.client.execute(
            f'cpg.graph.V.map(node=>node).toJson'
        )
        res_out = out['stdout'].split('String = ')[-1]
        return eval(res_out)

    #
    # Private Helpers
    #
    
    # TODO: for future we should reduce number of queries we are running for the speed. 
    # consider fetching multiple info in one query. 
    # for example: cpg.method.filter(node => node.lineNumber!=None&&node.lineNumberEnd!=None&&node.lineNumber!=node.lineNumberEnd).map(node=>(node.name, node.fullName, node.filename)).l
    
    def _get_functions_with_code(self):
        out = self._exec_list_cmd(
            'cpg.method.filter(node => node.lineNumber!=None&&node.lineNumberEnd!=None&&node.body.typeFullName!="<empty>").name.l' # Filter all the declarations
        )
        if not isinstance(out, (tuple, list)):
            return []

        out = list(out)
        # return set(self._filter_blacklisted(out))
        return set(out)

    @staticmethod
    def _filter_blacklisted(strings):
        blacklist = [
            "<", "+", "*", "(", ">", "JUMPOUT", "__builtin_unreachable"
        ]
        good_strings = []
        for string in strings:
            if not any(string.startswith(b) for b in blacklist):
                good_strings.append(string)

        return good_strings

    def _exec_list_cmd(self, raw_command):
        res = self.client.execute(f"{raw_command}")
        tuple_str = self._get_str_tuple(res, cmd=raw_command)
        return tuple_str

    def _exec_int_cmd(self, raw_command):
        res = self.client.execute(f"{raw_command}")
        int_val = self._get_str_int(res, cmd=raw_command)
        if int_val is None or not isinstance(int_val, int):
            return None

        return int_val
    
    def _get_str_int(self, req_res, cmd=None):
        if 'stdout' not in req_res:
            return None

        out = None
        try:
            raw_out = req_res['stdout']
            out = int(raw_out, 10)
        except Exception as e:
            l.warning(f"Error occurred doing JOERN eval for {self.bin_name} on {cmd} because {e}")

        return out

    def _get_str_tuple(self, req_res, cmd=None):
        if 'stdout' not in req_res:
            return None
        
        try:
            raw_out = req_res['stdout'].split("= List(")[-1]
            str_tuple = eval("(" + raw_out)
        except Exception as e:
            l.warning(f"Error occurred doing JOERN eval for {self.bin_name} on {cmd} because {e}")
            return None

        if type(str_tuple) == str:
            str_tuple = [str_tuple]

        return str_tuple

    def _dump_func_cfg(self, func_name):
        str_tuple = self._exec_list_cmd(f'cpg.method.filter(_.name=="{func_name}").dotCfg.l')
        try:
            graph = nx.nx_agraph.from_agraph(pg.AGraph(str_tuple[-1]))
        except Exception as e:
            l.warning(f"Error getting CFG from JOERN for {self.bin_name} on {func_name} as {e}")
            graph = None

        return graph

    def _list_parsing_regex(self, raw_command):
        res = self.client.execute(f"{raw_command}")
        if 'stdout' not in res:
            return None
        
        result_list = []
        raw_out = res['stdout'].split('] = List')[-1]

        # TODO set how many elements you need in each tuple for pattern_gen
        pattern = r'\(\s*(\d+L)\s*,\s*(\d+L)\s*,\s*"(.*?)"\s*,\s*(.*?)\s*\)'
        regex = re.compile(pattern)
        matches = regex.finditer(raw_out)
        for m in matches:
            elements = m.groups()
            result_list.append([(elem) for elem in elements])
        return result_list
        
      


import subprocess
from time import time, sleep
import os
import logging

from cpgqls_client import CPGQLSClient
import psutil

from . import JOERN_SERVER_PATH

l = logging.getLogger(__name__)


class JoernServer:
    def __init__(self, ip="localhost", port=9000):
        self.ip = ip
        self.port = port

    def __enter__(self):
        self.start()
        self.wait_for_server_start()
        return self

    def __exit__(self, *args, **kwargs):
        self.stop()

    def start(self):
        # always kill proc of the same port before starting
        joern_proc = self._find_joern_proc()
        if joern_proc:
            joern_proc.kill()

        # this spawns a child process as well
        proc = subprocess.Popen(
            f"{JOERN_SERVER_PATH} --server --server-port {self.port}".split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        sleep(3)
        if proc.poll() is not None:
            error_msg = proc.stderr.read().decode("utf-8")
            raise Exception(f"Unable to start the Joern server on port {self.port} because error: {error_msg}")

        return True

    def wait_for_server_start(self, timeout=30):
        start_time = time()
        success = False
        while time() - start_time < timeout:
            if self._find_joern_proc():
                try:
                    CPGQLSClient(f"{self.ip}:{self.port}").execute("val a = 1")
                except (ConnectionRefusedError, OSError):
                    pass
                else:
                    success = True
                    break

            sleep(0.5)

        if not success:
            l.critical(f"Was unable to start the JOERN server on port {self.port} before timeout...")
        else:
            sleep(0.5)

    def stop(self):
        joern_proc = self._find_joern_proc()
        if joern_proc:
            joern_proc.kill()
        else:
            l.critical(f"Unable to kill the Joern server on port {self.port} because it's dead our changed id")

        try:
            os.unlink("t.tmp")
        except Exception:
            pass

    def reboot(self):
        self.stop()
        self.start()
        self.wait_for_server_start()

    def _find_joern_proc(self):
        for proc in psutil.process_iter():
            try:
                cmd = " ".join(proc.cmdline())
            except Exception as e:
                continue

            if "java" in cmd and "joern" in cmd and f"--server-port {self.port}" in cmd:
                break
        else:
            proc = None

        return proc




from pathlib import Path
import importlib.resources
import subprocess
import json
import re

from .. import JOERN_SERVER_PATH
from .function import Function

SCALA_SCRIPT_PATH = Path(Path(str(importlib.resources.files("pyjoern"))) / "scala").absolute()
FAST_PARSER_SCRIPT = SCALA_SCRIPT_PATH / "FastParser.sc"
START_DELIM = "PYJOERN_DATA_START\n"
END_DELIM = "PYJOERN_DATA_END\n"


def _run_fast_parser_scala_script(source_path: Path) -> list[dict]:
    if not JOERN_SERVER_PATH.exists():
        raise FileNotFoundError(f"Joern server binary not found at {JOERN_SERVER_PATH}")
    if not FAST_PARSER_SCRIPT.exists():
        raise FileNotFoundError(f"Fast parser script not found at {FAST_PARSER_SCRIPT}")

    cmd = [
        str(JOERN_SERVER_PATH),
        "--script",
        str(FAST_PARSER_SCRIPT),
        "--param",
        f'target_dir={source_path}'
    ]
    proc = subprocess.run(cmd, capture_output=True, text=True)
    if proc.returncode != 0 or not proc.stdout:
        raise RuntimeError(f"Fast parser failed in script call. Stderr: {proc.stderr} | Stdout: {proc.stdout}")

    parsed_data = []
    found_jsons = re.findall(f"{START_DELIM}(.*?){END_DELIM}", proc.stdout, flags=re.DOTALL)
    for json_str in found_jsons:
        try:
            parsed_data.append(json.loads(json_str))
        except json.JSONDecodeError as e:
            raise RuntimeError(f"Fast parser script made non-json compliant output: {json_str[:30]}...")

    return parsed_data


def parse_source(source_path: Path) -> dict[str, Function] | dict[tuple[str, str], Function]:
    source_path = Path(source_path).absolute()
    if not source_path.exists():
        raise FileNotFoundError(f"Source file {source_path} does not exist!")

    data_dict = _run_fast_parser_scala_script(source_path)
    functions_by_name = Function.from_many(data_dict)
    if not source_path.is_dir():
        # remove file name from dict since they are all the same
        functions_by_name = {
            k[0]: v for k, v in functions_by_name.items()
        }

    return functions_by_name



import re
from typing import Dict, List
from collections import defaultdict
import logging

from .block import Block
from .statement import (
    UnsupportedStmt, UnknownStmt, Assignment, Compare, BinOp, Return, Call, Ternary, Nop
)

import networkx as nx

l = logging.getLogger(__name__)

STMT_MAP = {
    "assignment": (Assignment, None),
    "minus": (BinOp, BinOp.SUB),
    "plus": (BinOp, BinOp.ADD),
    "logicalAnd": (BinOp, BinOp.AND),
    "logicalOr": (BinOp, BinOp.OR),
    "equals": (Compare, Compare.EQ),
    "greaterEqualsThan": (Compare, Compare.GTE),
    "greaterThan": (Compare, Compare.GT),
    "lessEqualsThan": (Compare, Compare.LTE),
    "lessThan": (Compare, Compare.LT),
    "notEquals": (Compare, Compare.NE),
    "return": (Return, None),
    "call": (Call, None),
    "conditional": (Ternary, None),
}


def lift_graph(graph: nx.DiGraph):
    lifted_graph = nx.DiGraph()
    edges_by_node_addr: Dict[int, List[int]] = defaultdict(list)
    for in_node, out_node, _ in graph.edges(data=True):
        edges_by_node_addr[in_node].append(out_node)

    addr_uses = defaultdict(int)
    new_blocks = {
        node: lift_block(node, graph, addr_dict=addr_uses) for node in graph.nodes
    }

    for node_name, new_node in new_blocks.items():
        for out_node in edges_by_node_addr[node_name]:
            new_out_block = new_blocks[out_node]
            lifted_graph.add_edge(new_node, new_out_block)

    return lifted_graph


def lift_block(node, graph, addr_dict=None):
    node_data = graph._node[node].get('label', None)
    if not node_data:
        new_addr = 0 - int(node, 10)
        new_idx = addr_dict[new_addr]
        addr_dict[new_addr] += 1
        return Block(0 - int(node, 10), statements=[Nop("", Nop.NOP)], idx=new_idx)

    statements = []
    for raw_stmt in node_data.split("\n"):
        try:
            lifted_stmt = lift_statement(raw_stmt)
        except Exception as e:
            l.debug(f"Encountered error while lifting: {e}")
            lifted_stmt = UnsupportedStmt(raw_stmt)

        statements.append(lifted_stmt)

    new_addr = statements[0].source_line_number
    new_idx = addr_dict[new_addr]
    addr_dict[new_addr] += 1
    return Block(new_addr, statements=statements, idx=new_idx)


def lift_statement(raw_data: str):
    source_line_number = re.findall("<SUB>([0-9]+)</SUB>", raw_data)
    if source_line_number:
        source_line_number = int(source_line_number[0], 10)
    else:
        l.debug(f"Failed to parse line number of node!")
        source_line_number = None
    line_num_kwarg = {"source_line_number": source_line_number}

    _t = "(".join(raw_data.split("(")[1:])
    raw_stmt = "".join(_t.split(")<SUB>")[0:-1]).replace(" ", "")

    # statement
    is_stmt = True
    stmt_type = None
    if raw_stmt.startswith("&lt;operator&gt;."):
        # &lt;operator&gt;.lessEqualsThan,v0 &lt;= line_buf
        t = raw_stmt.split("&lt;operator&gt;.")[1]
        t = t.split(",")
        stmt_type = t[0]
    elif raw_stmt.startswith("RETURN"):
        t = raw_stmt.split("RETURN")[1]
        t = t.split(",")
        stmt_type = "return"
    elif raw_stmt.startswith("METHOD_RETURN"):
        return Nop(raw_stmt, Nop.FUNC_END, **line_num_kwarg)
    elif raw_stmt.startswith("METHOD"):
        return Nop(raw_stmt, Nop.FUNC_START, **line_num_kwarg)
    else:
        is_stmt = False

    if is_stmt:
        # [lessEqualsThan], [v0 &lt;= line_buf],...
        raw_ops = ",".join(t[1:])
        # v0 &lt;= line_buf ...


        (jil_stmt_cls, sub_type) = STMT_MAP.get(stmt_type, (UnsupportedStmt, None))
        ops = []
        if jil_stmt_cls != UnsupportedStmt:
            sep = jil_stmt_cls.op_seperator(sub_type)
            if sep:
                out = raw_ops.split(sep)
                if not out[0]:
                    ops = [f"{sep}".join(out[1:])]
                else:
                    ops = out[0:1] + [f"{sep}".join(out[1:])]

        # specific parsers
        if jil_stmt_cls in (Compare, BinOp):
            return jil_stmt_cls(t, sub_type, *ops, **line_num_kwarg)
        # SPECIAL CASE: ternary
        elif jil_stmt_cls == Ternary:
            t = raw_ops.split("?")
            cond = t[0]
            t = "?".join(t)
            t.split(":")
            true = t[0]
            false = t[1]
            return Ternary(raw_stmt, cond, true, false, **line_num_kwarg)
        else:
            return jil_stmt_cls(t, *ops, **line_num_kwarg)
    else:
        t = raw_stmt.replace(" ", "")

        # check if call
        try:
            t_no_comma = t.split(",")
            is_call = t_no_comma and t_no_comma[0] == t_no_comma[1].split("(")[0]
        except (KeyError, IndexError):
            is_call = False
        if is_call:
            raw_call = t_no_comma[1]
            func = raw_call.split("(")[0]
            raw_args = "(".join(raw_call.split("(")[1:])
            raw_args = ")".join(raw_args.split(")")[:-1])
            args = raw_args.split(",")
            return Call(raw_stmt, func, args, **line_num_kwarg)

    return UnsupportedStmt(raw_stmt, **line_num_kwarg)



from pathlib import Path
from tempfile import TemporaryDirectory
from subprocess import run
import logging

import networkx as nx
import pygraphviz as pg
from cfgutils.transformations import to_supergraph

from .jil.lifter import lift_graph
from .. import JOERN_PARSE_PATH, JOERN_EXPORT_PATH
from ..utils import WorkDirContext

_l = logging.getLogger(__name__)


def fast_cfgs_from_source(filepath: Path, lift_cfgs=True, supergraph=True, timeout=120):
    filepath = Path(filepath).absolute()
    cfgs = {}
    with TemporaryDirectory() as tmpdir:
        with WorkDirContext(tmpdir):
            # run joern-parser which will output a cpg in the same dir by the
            # filename of "cpg.bin"
            ret = run(f"{JOERN_PARSE_PATH} {filepath}".split(), capture_output=True, timeout=timeout)
            if ret.returncode != 0:
                _l.warning("Joern parse failed, stopping CFG extraction")
                return None

            # extras the cfgs into out_dir in same directory
            ret = run(f"{JOERN_EXPORT_PATH} --repr cfg --out out_dir".split(), capture_output=True, timeout=timeout)
            if ret.returncode != 0:
                _l.warning("Joern Export failed, stopping CFG extraction")
                return None

            out_dir = Path("./out_dir")
            cfg_files = list(out_dir.rglob("*.dot"))
            for cfg_file in cfg_files:
                cfg = cfg_from_dotfile(cfg_file.absolute())
                if not cfg or not len(cfg.nodes):
                    continue

                if cfg.name in cfgs and len(cfg.nodes) < len(cfgs[cfg.name].nodes):
                    continue

                cfgs[cfg.name] = nx.DiGraph(cfg)

    normalized_cfgs = {}
    for cfg_name, cfg in cfgs.items():
        normalized_cfgs[cfg_name] = normalize_cfg(cfg, lift_cfg=lift_cfgs, supergraph=supergraph)

    return normalized_cfgs


def normalize_cfg(cfg: nx.DiGraph, lift_cfg=True, supergraph=True):
    if lift_cfg:
        jil_cfg = lift_graph(cfg)
        jil_cfg.name = cfg.name
        jil_cfg = to_supergraph(jil_cfg) if supergraph else jil_cfg
        cfg = jil_cfg

    node_attrs = {}
    edge_attrs = {}
    for node in cfg.nodes:
        node_attrs[node] = {'node': node}
    for edgd in cfg.edges:
        edge_attrs[edgd] = {'src': edgd[0], 'dst': edgd[1]}

    nx.set_node_attributes(cfg, node_attrs)
    nx.set_edge_attributes(cfg, edge_attrs)
    return cfg


def parse_dot_cfg_string(dotcfg_string):
    try:
        graph = nx.nx_agraph.from_agraph(pg.AGraph(dotcfg_string))
    except Exception as e:
        _l.critical(f"Failed to parse CFG from dot string: {e}")
        graph = None

    return graph


def cfg_from_dotfile(filepath: Path):
    filepath = Path(filepath).expanduser().absolute()
    with open(filepath, "r") as fp:
        data = fp.read()

    return parse_dot_cfg_string(data)
